from logging import basicConfig, getLogger
from pathlib import Path
from subprocess import DEVNULL, run

from adjudicator import Rule, RulesEngine
from equilibrium import Resource, ResourceContext, Service
from rich.logging import RichHandler
from typer import Argument, Typer

from helsing.scm.tfcodegen import generate_terraform_code

logger = getLogger(__name__)
app = Typer(pretty_exceptions_enable=False)


def get_resources() -> list[type[Resource.Spec]]:
    """
    Return the resources to use for generating Terraform code.
    """

    import helsing.scm.resources
    import helsing.scm.tfcodegen

    return [
        *helsing.scm.resources.get_resources(),
        *helsing.scm.tfcodegen.get_resources(),
    ]


def get_services() -> list[Service]:
    """
    Return the services to use for generating Terraform code.
    """

    import helsing.scm.tfcodegen

    return [
        *helsing.scm.tfcodegen.get_services(),
    ]


def get_rules() -> list[Rule]:
    """
    Return the rules to use for generating Terraform code.
    """

    import helsing.scm.tfcodegen

    return [
        *helsing.scm.tfcodegen.get_rules(),
    ]


@app.command()
def main(
    output_dir: Path = Path("generated"),
    manifest_files: list[Path] = Argument(...),
) -> None:
    """
    Secure Cloud Manager generates Terraform code from a declarative specification.
    """

    basicConfig(
        level="INFO",
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(rich_tracebacks=True)],
    )

    context = ResourceContext.create(ResourceContext.InMemoryBackend())
    for resource_type in get_resources():
        logger.info("Register resource type '%s'", resource_type.TYPE)
        context.resource_types.register(resource_type)
    for service_type in get_services():
        logger.info("Register service '%s' for resource type '%s'", service_type.SERVICE_ID, service_type.RESOURCE_TYPE)
        context.services.register(service_type)

    engine = RulesEngine(get_rules(), [context])
    engine.hashsupport.register(Resource, lambda r: hash(r.uri))

    for path in manifest_files:
        logger.info("Loading manifest '%s'", path)
        context.load_manifest(path)

    logger.info("Generating Terraform code in '%s'", output_dir)

    with engine.as_current():
        generate_terraform_code(context=context, directory=output_dir)

    logger.info("Running 'terraform fmt' on directory '%s'", output_dir)
    exit(run(["terraform", "fmt", "--recursive", output_dir], stdout=DEVNULL).returncode)


if __name__ == "__main__":
    app()



"""
Implements the configuration spec for AWS accounts.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Literal, Sequence

from equilibrium import Resource, ResourceContext

from helsing.scm import DOMAIN, GIT_REPOSITORY_URL
from helsing.scm.utils.develop import get_develop_root
from helsing.scm.utils.types import FrozenDict, HashableMapping

# TODO(@NiklasRosenstein): Databind will deserialize Sequence as a list; but we'd actually
#       want that to be an immutable sequence type. We should maybe update databind to support this.
JsonValue = str | int | float | bool | FrozenDict[str, Any] | Sequence[Any]


@dataclass(frozen=True)
class ScmModule:
    """
    Reference a Terraform module that is available as part of the SCM, either referenced from the local filesystem
    (only possible when SCM is installed in Python develop mode) or from the GitHub repository.
    """

    #: The name of the Terraform module in the SCM `modules/` directory.
    name: str

    #: Where to source the SCM module from.
    source: Literal["Local", "Remote"] = "Remote"

    #: The Git ref to use when sourcing the module from the SCM repository. This is only used when `source` is `Remote`.
    ref: str | None = None

    #: Additional static input variables to set when the module is instantiated.
    additionalInputVariables: HashableMapping[str, JsonValue] = field(default_factory=FrozenDict)

    type: Literal["ScmModule"] = "ScmModule"

    def as_terraform_module(self) -> TerraformModule:
        """
        Convert this SCM module to a Terraform module.
        """

        match self.source:
            case "Local":
                source = str(get_develop_root() / "modules" / self.name)
            case "Remote":
                source = f"git::{GIT_REPOSITORY_URL}//modules/{self.name}"
                if self.ref is not None:
                    source += f"?ref={self.ref}"
            case _:
                raise ValueError(f"Invalid source {self.source!r}")

        return TerraformModule(
            source=source,
            version=None,
            additionalInputVariables=self.additionalInputVariables,
        )


@dataclass(frozen=True)
class TerraformModule:
    """
    Reference a Terraform module that is available from the Terraform registry or from a Git repository.
    """

    #: The Terraform module source. This must be the full string that you would also set in the `source`
    #: attribute of the `module {}` block.
    source: str

    #: The version of the Terraform module to use.
    version: str | None = None

    #: Additional static input variables to set when the module is instantiated.
    additionalInputVariables: HashableMapping[str, JsonValue] = field(default_factory=FrozenDict)

    type: Literal["Terraform"] = "Terraform"

    def as_terraform_module(self) -> TerraformModule:
        return self


@dataclass(frozen=True)
class AwsTerraformCodegenSettings(
    Resource.Spec,
    apiVersion=f"{DOMAIN}/v1alpha1",
    kind="AwsTerraformCodegenSettings",
    namespaced=False,
):
    """
    Contains settings for Terraform code generation related to AWS account creation.
    """

    #: The Terraform module that is used to instantiate the AWS account. The module must have, at minimum, the
    #: following inputs varibales:
    #:
    #: - `account_name`: The name of the account.
    #: - `account_email`: The email address of the account.
    #: - `organizational_unit`: The organizational unit of the account (full name, not the ID).
    #:
    #: The module must output:
    #:
    #: - `account_id`: The AWS account ID.
    #: - `access_key`: The AWS access key.
    #: - `secret_key`: The AWS secret key.
    #: - `admin_role_arn`: The AWS administrator role ARN in the account.
    terraformModule: ScmModule | TerraformModule

    @staticmethod
    def get(context: ResourceContext) -> Resource[AwsTerraformCodegenSettings]:
        resources = context.resources.list(AwsTerraformCodegenSettings)
        if len(resources) == 0:
            raise RuntimeError("No AwsTerraformCodegenSettings resource found.")
        if len(resources) > 1:
            raise RuntimeError("Multiple AwsTerraformCodegenSettings resources found.")
        return resources[0]


def get_resources() -> list[type[Resource.Spec]]:
    return [AwsTerraformCodegenSettings]



from abc import abstractmethod
from typing import Any

from equilibrium import Resource, Service

from helsing.scm import DOMAIN
from helsing.scm.resources.v1alpha.TerraformWorkspace import ProviderFrom, Variable


class TerraformAccountService(Service, serviceId=f"{DOMAIN}/v1alpha1/TerraformAccountService"):
    """
    This service provides an API to retrieve Terraform workspace configuration for a resource that resembles a
    Cloud "account". It is used for resources referenced in #ApiDrivenWorkflow.providersFrom to expose credentials
    in a workspace.

    Using this service, a Terraform workspace can be configured to use the credentials of a resource for which
    an implementation of this service exists without the Terraform workspace code generation to having to understand
    the resource type. This allows us to extend the set of Cloud account resources that a Terraform workspace can
    consume credentials from in the future.
    """

    @abstractmethod
    def get_variables(self, resource: Resource[Any]) -> list[Variable]:
        """
        Return a list of variables that should be assigned to the Terraform workspace, which are derived from
        the specified resource.
        """

    @abstractmethod
    def get_provider_initialization(self, provider_from: ProviderFrom, resource: Resource[Any]) -> str:
        """
        Return the code that will be put into an API-driven Terrafork workspace to initialize the provider for
        the specified resource.
        """



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Literal

from equilibrium import Resource

from helsing.scm import DOMAIN


@dataclass(frozen=True)
class ProviderFrom:
    """
    Represents a reference to another resource that must be supported as a source for credentials and the generation
    of initialization code for a Terraform provider.
    """

    #: The reference to the resource.
    accountRef: Resource.URI

    #: The version of the provider to use. If this is not set, no version will be set in the generated code.
    #: The provider type is fixed by the type of resource referenced with #accountRef.
    providerVersion: str | None = None

    #: An optional alias for the provider.
    alias: str | None = None


@dataclass(frozen=True)
class Variable:
    """
    Represents a Terraform variable that will be assigned to a Terraform workspace.
    """

    #: The name of the variable.
    name: str

    #: The value of the variable.
    value: str | Expression

    #: An optional description of the variable.
    description: str | None = None

    #: Whether the variable is sensitive.
    sensitive: bool = False

    #: The category of the variable.
    category: Literal["Terraform", "Env"] = "Terraform"

    #: Whether the value is in HCL format.
    hcl: bool = False


@dataclass
class Expression:
    """
    Represents an expression of Terraform code.
    """

    #: The expression.
    value: str

    type: Literal["Expression"] = "Expression"


@dataclass(frozen=True)
class ApiDrivenWorkflow:
    """
    Represents API-driven Terraform Workspace run workflow as described here:

    https://developer.hashicorp.com/terraform/cloud-docs/run/api
    """

    #: A reference to a Terraform module. This can be a module from a a Git repository. In Terraform, a Git repository
    #: must be prefixed with "git::", but here we can determine that a Git repository is referenced by a simpler URL
    #: format, such as simply `git@github:my-org/my-repo`.
    moduleSource: str

    #: The version of the module to use. If a Git module is specified with #moduleSource, this must be a Git reference
    #: or commit and will be appended to the module source URL with a `?ref=` query parameter.
    moduleVersion: str | None = None

    #: A list of resources from which to source provider credentials and initialization code.
    providersFrom: list[ProviderFrom] = field(default_factory=list)

    #: A list of variables to assign to the workspace and to forward as inputs to the Terraform module. These variables
    #: should not conflict with variables specified in the #TerraformWorkspace.variables list.
    variables: list[Variable] = field(default_factory=list)

    type: Literal["ApiDriven"] = "ApiDriven"


@dataclass(frozen=True)
class TerraformWorkspace(Resource.Spec, apiVersion=f"{DOMAIN}/v1alpha1", kind="TerraformWorkspace", namespaced=False):
    """
    Represents a Terraorm workspace which may be given access to an AWS account.
    """

    #: The name of the Terraform workspace.
    workspaceName: str

    #: The type of workflow that the workspace is using to manage infrastructure,
    workflow: ApiDrivenWorkflow

    #: A list of variables that will be assigned to the Terraform workspace. The variable will not be considerted
    #: further. For API-driven workflows, variables that should also be forwarded to the instantiated Terraform module,
    #: they must be specified in the #ApiDrivenWorkflow.variables list.
    variables: list[Variable] = field(default_factory=list)



from adjudicator import Rule
from equilibrium import Resource, Service

from . import account_service, create_account, settings


def get_rules() -> list[Rule]:
    return [
        *create_account.get_rules(),
    ]


def get_resources() -> list[type[Resource.Spec]]:
    return [
        *settings.get_resources(),
    ]


def get_services() -> list[Service]:
    return [
        account_service.AwsAccountService(),
    ]



import re
from dataclasses import dataclass

from equilibrium import Resource

from helsing.scm import DOMAIN

ACCOUNT_NAME_REGEX = r"^[a-zA-Z0-9][a-zA-Z0-9._-]*$"


@dataclass(frozen=True)
class AwsAccount(Resource.Spec, apiVersion=f"{DOMAIN}/v1alpha1", kind="AwsAccount", namespaced=False):
    """
    Represents an AWS account.
    """

    #: The name of the AWS account in the AWS console.
    accountName: str

    #: The email address associated with the AWS account.
    accountEmail: str

    #: The default region for the AWS account.
    defaultRegion: str

    #: The organizational unit that the AWS account belongs to. This must be a unique name in the
    #: management account. If not specified, the account will be created in the Root OU.
    organizationalUnit: str | None = None

    def __post_init__(self) -> None:
        assert re.match(
            ACCOUNT_NAME_REGEX, self.accountName
        ), f"Invalid account name: {self.accountName!r}, must match {ACCOUNT_NAME_REGEX}"



"""
Implements generating Terraform code.
"""

from logging import getLogger
from pathlib import Path

from adjudicator import Rule, get
from equilibrium import Resource, ResourceContext, Service

from helsing.scm.resources.v1alpha.AwsAccount import AwsAccount
from helsing.scm.resources.v1alpha.TerraformWorkspace import TerraformWorkspace
from helsing.scm.tfcodegen import aws, tfe

logger = getLogger(__name__)


def get_rules() -> list[Rule]:
    """
    Return the rules to use for generating Terraform code.
    """

    from . import aws, tfe

    return [
        *aws.get_rules(),
        *tfe.get_rules(),
    ]


def get_services() -> list[Service]:
    """
    Return the services to use for generating Terraform code.
    """

    from . import aws

    return [
        *aws.get_services(),
    ]


def get_resources() -> list[type[Resource.Spec]]:
    """
    Return the resources to use for generating Terraform code.
    """

    from . import aws, tfe

    return [
        *aws.get_resources(),
        *tfe.get_resources(),
    ]


def generate_terraform_code(context: ResourceContext, directory: Path) -> None:
    """
    Generate Terraform code for the given resources in the given directory.
    """

    for aws_account in context.resources.list(AwsAccount):
        aws_code = get(
            aws.create_account.AwsAccountCreationCode,
            {Resource[AwsAccount]: aws_account},
        )
        path = directory / f"aws-account-{aws_account.metadata.name}.tf"
        logger.info("Write '%s'", path)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(aws_code.value)

    for tfe_workspace in context.resources.list(TerraformWorkspace):
        tfe_code = get(
            tfe.create_workspace.TerraformWorkspaceCreationCode,
            {Resource[TerraformWorkspace]: tfe_workspace},
        )
        path = directory / f"tfe-workspace-{tfe_workspace.metadata.name}.tf"
        logger.info("Write '%s'", path)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(tfe_code.value)



from pathlib import Path


def get_develop_root() -> Path:
    """
    Returns the path to the root directory of the SCM repository. If SCM is installed as a package instead of in
    develop mode, this function will raise a #ValueError exception.

    The function uses the resolved path of the `helsing.scm` module and checks if it lives in a directory that
    corresponds to the structure of the SCM Git repository.
    """

    from helsing import scm

    source_root = Path(scm.__file__).resolve().parent.parent.parent
    if source_root.name != "src":
        raise ValueError(f"SCM is not installed in develop mode: {source_root}")

    readme = source_root.parent / "README.md"
    if not readme.is_file():
        raise ValueError(f"SCM is not installed in develop mode: {source_root}")

    return source_root.parent


if __name__ == "__main__":
    print(get_develop_root())



import inspect
from textwrap import dedent
from typing import Literal

from mako.template import Template  # type: ignore


def template(template: str, dedent_mode: Literal["dedent"] | None = None, /, **kwargs: object) -> str:
    """
    Render a template string using Mako.

    :param template: The Mako template string.
    :param dedent: Whether to dedent the template string before rendering.
    :param kwargs: The template variables.
    """

    if dedent_mode == "dedent":
        template = dedent(template)
    filename = inspect.stack()[1].filename
    value = Template(template, filename=filename).render(**kwargs)
    assert isinstance(value, str)
    return value



from typing import Hashable, Mapping, TypeVar

from frozendict import frozendict

__all__ = ["HashableMapping", "FrozenDict"]

K_Hashable = TypeVar("K_Hashable", bound=Hashable)
V_Hashable = TypeVar("V_Hashable", bound=Hashable)


class HashableMapping(Mapping[K_Hashable, V_Hashable], Hashable):
    """
    A hashable mapping.
    """


class FrozenDict(frozendict[K_Hashable, V_Hashable], HashableMapping[K_Hashable, V_Hashable]):
    """
    A concrete implementation of a hashable mapping.
    """

    pass



"""
Implements the configuration spec for AWS accounts.
"""

from __future__ import annotations

from dataclasses import dataclass

from equilibrium import Resource, ResourceContext

from helsing.scm import DOMAIN
from helsing.scm.tfcodegen.aws.settings import ScmModule, TerraformModule


@dataclass(frozen=True)
class TerraformWorkspaceCodegenSettings(
    Resource.Spec,
    apiVersion=f"{DOMAIN}/v1alpha1",
    kind="TerraformWorkspaceCodegenSettings",
    namespaced=False,
):
    """
    Contains settings for Terraform code generation related to TFE workspace creation.
    """

    #: The Terraform module that is used to instantiate the AWS account. The module must have, at minimum, the
    #: following inputs varibales:
    #:
    #: - `variables`: A map of variable configurations that are assigned to the workspace. See #Variable.
    #: - `configuration`: A string that represents the Terraform code to execute as part of the workspace.
    terraformModule: ScmModule | TerraformModule

    @staticmethod
    def get(context: ResourceContext) -> Resource[TerraformWorkspaceCodegenSettings]:
        resources = context.resources.list(TerraformWorkspaceCodegenSettings)
        if len(resources) == 0:
            raise RuntimeError("No TerraformWorkspaceCodegenSettings resource found.")
        if len(resources) > 1:
            raise RuntimeError("Multiple TerraformWorkspaceCodegenSettings resources found.")
        return resources[0]


def get_resources() -> list[type[Resource.Spec]]:
    return [TerraformWorkspaceCodegenSettings]



"""
Implements the Terraform code generation for AWS accounts.
"""

import json
from dataclasses import dataclass
from string import Template

from adjudicator import Rule, collect_rules, rule
from equilibrium import Resource, ResourceContext

from helsing.scm.resources.v1alpha.AwsAccount import AwsAccount
from helsing.scm.utils.templating import template

from .settings import AwsTerraformCodegenSettings


@dataclass(frozen=True)
class AwsAccountCreationCode:
    #: The generated code.
    value: str

    #: The address of the AWS module resource that was generated in the code.
    module_address: str


def render_template_string(aws_account: Resource[AwsAccount], template: str) -> str:
    return Template(template).substitute(identifier=aws_account.metadata.name)


@rule
def create_aws_account(context: ResourceContext, aws_account: Resource[AwsAccount]) -> AwsAccountCreationCode:
    """
    Generate Terraform code that creates an AWS account and stores its credentials in Vault.
    """

    config = AwsTerraformCodegenSettings.get(context)
    module = config.spec.terraformModule.as_terraform_module()

    normalized_name = aws_account.metadata.name.replace("-", "_")

    code = template(
        """
        module "aws_account_${normalized_name}" {
            source = "${module.source}"
            % if module.version:
            version = "${module.version}"
            % endif
            account_name = "${aws_account.spec.accountName}"
            account_email = "${aws_account.spec.accountEmail}"
            organizational_unit = "${aws_account.spec.organizationalUnit}"
            % for key, value in module.additionalInputVariables.items():
            ${key} = ${dumps(render_template_string(aws_account, value)) if isinstance(value, str) else dumps(value)}
            % endfor
        }
        """,
        "dedent",
        dumps=json.dumps,
        render_template_string=render_template_string,
        module=module,
        aws_account=aws_account,
        config=config,
        normalized_name=normalized_name,
    )

    return AwsAccountCreationCode(value=code, module_address=f"module.aws_account_{normalized_name}")


def get_rules() -> list[Rule]:
    return collect_rules()



import json
from typing import Any

from adjudicator import get
from equilibrium import Resource

from helsing.scm.resources.v1alpha.AwsAccount import AwsAccount
from helsing.scm.resources.v1alpha.TerraformWorkspace import Expression, ProviderFrom, Variable
from helsing.scm.tfcodegen.aws.create_account import AwsAccountCreationCode
from helsing.scm.tfcodegen.tfe.account_service import TerraformAccountService
from helsing.scm.utils.templating import template


class AwsAccountService(TerraformAccountService, resourceType=AwsAccount):
    """
    The implementation for exposing AWS account credentials to a Terraform workspace.
    """

    def get_variables(self, resource: Resource[object]) -> list[Variable]:
        code = get(AwsAccountCreationCode, {Resource[AwsAccount]: resource.into(AwsAccount)})
        return [
            Variable(
                name="aws_account_id",
                value=Expression(f"{code.module_address}.account_id"),
                category="Terraform",
                hcl=False,
                sensitive=False,
            ),
            Variable(
                name="aws_assume_role_arn",
                value=Expression(f"{code.module_address}.admin_role_arn"),
                category="Terraform",
                hcl=False,
                sensitive=False,
            ),
            Variable(
                name="AWS_ACCESS_KEY_ID",
                value=Expression(f"{code.module_address}.access_key"),
                category="Env",
                hcl=False,
                sensitive=True,
            ),
            Variable(
                name="AWS_SECRET_ACCESS_KEY",
                value=Expression(f"{code.module_address}.secret_key"),
                category="Env",
                hcl=False,
                sensitive=True,
            ),
        ]

    def get_provider_initialization(self, provider_from: ProviderFrom, resource: Resource[Any]) -> str:
        aws_account = resource.into(AwsAccount)
        return template(
            """
            provider "aws" {
                % if provider_from.alias:
                alias = ${dumps(provider_from.alias) or "null"}
                % endif
                region = "${aws_account.spec.defaultRegion}"
                allowed_account_ids = [var.aws_account_id]
                assume_role {
                    role_arn = var.aws_assume_role_arn
                }
            }
            """,
            "dedent",
            dumps=json.dumps,
            aws_account=aws_account,
            provider_from=provider_from,
        )



from adjudicator import Rule
from equilibrium import Resource

from . import create_workspace, settings


def get_resources() -> list[type[Resource.Spec]]:
    return [
        *settings.get_resources(),
    ]


def get_rules() -> list[Rule]:
    return [
        *create_workspace.get_rules(),
    ]



import json
from dataclasses import dataclass
from textwrap import indent

from adjudicator import Rule, collect_rules, rule
from equilibrium import Resource, ResourceContext

from helsing.scm import __version__
from helsing.scm.resources.v1alpha.TerraformWorkspace import TerraformWorkspace, Variable
from helsing.scm.tfcodegen.tfe.account_service import TerraformAccountService
from helsing.scm.tfcodegen.tfe.settings import TerraformWorkspaceCodegenSettings
from helsing.scm.utils.templating import template


@dataclass(frozen=True)
class TerraformCodegenConfig:
    """
    Settings for generating code related to the creation of Terraform workspaces.
    """

    #: The module that is used to create the Terraform workspace.
    workspace_module: str = f"git::https://github.com/helsing-ai/scm/modules/tfe_workspace?ref={__version__}"


@dataclass(frozen=True)
class TerraformWorkspaceCreationCode:
    """
    Represents the code that defines a Terraform workspace.
    """

    value: str


@rule
def create_terraform_workspace(
    context: ResourceContext,
    workspace: Resource[TerraformWorkspace],
) -> TerraformWorkspaceCreationCode:
    """
    Generate Terraform code that creates a Terraform workspace using the "tfe" provider.
    """

    settings = TerraformWorkspaceCodegenSettings.get(context)
    terraform_module = settings.spec.terraformModule.as_terraform_module()

    normalized_name = workspace.metadata.name.replace("-", "_")

    variables = list(workspace.spec.variables)
    provider_variables: list[Variable] = []
    provider_snippets: list[str] = []

    # Collect variables and provider initialization code snippets for the resources that the workspace depends on.
    for provider_from in workspace.spec.workflow.providersFrom:
        account = context.resources.get(provider_from.accountRef)
        service: TerraformAccountService = context.services.get(
            provider_from.accountRef.type, TerraformAccountService  # type: ignore[type-abstract]
        )
        provider_variables += service.get_variables(account)
        provider_snippets.append(service.get_provider_initialization(provider_from, account))

    code = template(
        """
        module "workspace_${normalized_name}" {
            source = "${module.source}"
            % if module.version:
            version = "${module.version}"
            % endif
            name = "${workspace.spec.workspaceName}"
            variables = {
            % for variable in variables + provider_variables:
                ${variable.name} = {
                    value = ${dumps(variable.value) if isinstance(variable.value, str) else variable.value.value}
                    % if variable.description:
                    description = "${variable.description}"
                    % endif
                    category = "${variable.category.lower()}"
                    % if variable.sensitive:
                    sensitive = true
                    % endif
                    % if variable.hcl:
                    hcl = true
                    % endif
                }
            % endfor
            }
            configuration = <<-EOF
                % for variable in variables + provider_variables:
                % if variable.category == "Terraform":
                variable "${variable.name}" {${ " sensitive = true " if variable.sensitive else "" }}
                % endif
                % endfor
                % for provider_snippet in provider_snippets:
                ${indent(provider_snippet, " " * 8)}
                % endfor
                module "main" {
                    source = "${workspace.spec.workflow.moduleSource}"
                    % if workspace.spec.workflow.moduleVersion:
                    version = "${workspace.spec.workflow.moduleVersion}"
                    % endif
                    % for variable in variables:
                    ${variable.name} = var.${variable.name}
                    % endfor
                }

                # TODO: Add individual outputs here / but how do we know which?
                output "module" {
                    value = module.main
                }
            EOF
            % for key, value in module.additionalInputVariables.items():
            ${key} = ${dumps(render_template_string(aws_account, value)) if isinstance(value, str) else dumps(value)}
            % endfor
        }
        """,
        "dedent",
        dumps=json.dumps,
        indent=indent,
        normalized_name=normalized_name,
        workspace=workspace,
        variables=variables,
        provider_variables=provider_variables,
        module=terraform_module,
        provider_snippets=provider_snippets,
    )

    return TerraformWorkspaceCreationCode(value=code)


def get_rules() -> list[Rule]:
    return collect_rules()



from equilibrium import Resource

from .AwsAccount import AwsAccount
from .TerraformWorkspace import TerraformWorkspace


def get_resources() -> list[type[Resource.Spec]]:
    return [
        AwsAccount,
        TerraformWorkspace,
    ]
